"""
Module defining the Loan class.
"""
import pathlib as pl
import numbers
import json
from typing import Optional
from dataclasses import dataclass

import pandas as pd
import voluptuous as vt

from . import constants as cs
from . import helpers as hp


@dataclass
class Loan:
    """
    Represents a loan.
    Attributes are

    - ``code``: code name for the loan; defaults to a
      timestamp-based code
    - ``kind``: kind of loan; 'amortized' or 'interest only'
    - ``principal``: amount of loan, that is, the principal
    - ``interest_rate``: nominal annual interest rate
      (not as a percentage), e.g. 0.1 for 10%
    - ``payment_freq``: payments frequency;
      one of the keys of :const:`NUM_BY_FREQ`, e.g. 'monthly'
    - ``compounding_freq``: compounding frequency;
      one of the keys of :const:`NUM_BY_FREQ`, e.g. 'monthly';
      defaults to ``payment_freq``
    - ``num_payments``: number of payments in the loan
      term
    - ``fee``: loan fee
    - ``first_payment_date``: (YYYY-MM-DD date string) date of first loan
      payment

    Validation happens automatically after creating a Loan instance, so no bad Loans
    can be created :)
    """

    code: str
    kind: str
    principal: float
    interest_rate: float
    payment_freq: str
    num_payments: int
    fee: float
    first_payment_date: str
    compounding_freq: Optional[str] = None  # defaults to payment_freq

    @staticmethod
    def validate(params: dict) -> dict:
        """
        Return the given parameters if they pertain to a valid Loan.
        Otherwise, raise a Voluptuous Invalid error.
        """

        def check_kind(value):
            kinds = ["amortized", "interest only"]
            if value in kinds:
                return value
            raise vt.Invalid("Kind must be one on {}".format(kinds))

        def check_pos(value):
            if isinstance(value, numbers.Number) and value > 0:
                return value
            raise vt.Invalid("Not a positive number")

        def check_nneg(value):
            if isinstance(value, numbers.Number) and value >= 0:
                return value
            raise vt.Invalid("Not a nonnegative number")

        def check_pos_int(value):
            if isinstance(value, int) and value > 0:
                return value
            raise vt.Invalid("Not a positive integer")

        def check_freq(value):
            if value in cs.NUM_BY_FREQ:
                return value
            raise vt.Invalid(f"Frequncy must be one of {cs.NUM_BY_FREQ.keys()}")

        def check_date(value, fmt="%Y-%m-%d"):
            err = vt.Invalid(f"Not a datestring of the form {fmt}")
            try:
                if value is not None and pd.to_datetime(value, format=fmt):
                    return value
                raise err
            except TypeError:
                raise err

        schema = vt.Schema(
            {
                "code": str,
                "kind": check_kind,
                "principal": check_pos,
                "interest_rate": check_nneg,
                "payment_freq": check_freq,
                "num_payments": check_pos_int,
                "first_payment_date": check_date,
                "fee": check_nneg,
                vt.Optional("compounding_freq"): check_freq,
            },
            required=True,
        )

        return schema(params)

    def __post_init__(self):
        # Set some defaults
        if self.compounding_freq is None:
            self.compounding_freq = self.payment_freq

        # Validate
        attrs = self.__dict__
        Loan.validate(attrs)

    def copy(self) -> "Loan":
        """
        Return a copy of this Loan.
        """
        return Loan(**self.__dict__)

    def summarize(self, decimals: int = 2) -> dict:
        """
        Return the result of :func:`helpers.summarize_amortized_loan`
        is this Loan is amortized or
        :func:`helpers.summarize_interest_only_loan` if this Loan is
        interest-only.
        """
        result = {}
        if self.kind == "amortized":
            result = hp.summarize_amortized_loan(
                self.principal,
                self.interest_rate,
                self.compounding_freq,
                self.payment_freq,
                self.num_payments,
                self.fee,
                self.first_payment_date,
                decimals=decimals,
            )
        elif self.kind == "interest only":
            result = hp.summarize_interest_only_loan(
                self.principal,
                self.interest_rate,
                self.payment_freq,
                self.num_payments,
                self.fee,
                self.first_payment_date,
                decimals=decimals,
            )

        return result


def read_loan(path: pl.PosixPath) -> "Loan":
    """
    Given a path to a JSON file encoding the attributes
    of a Loan, read the file, and return the corresponding Loan instance.
    Extra keys in the JSON file will be ignored.
    """
    with pl.Path(path).open() as src:
        params = json.load(src)

    # Prune params
    pruned_params = {}
    for k in params:
        if k in Loan.__dataclass_fields__:
            pruned_params[k] = params[k]

    return Loan(**pruned_params)
